﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

namespace FourInARow
{
    class Drawer
    {
        #region Colors

        private Pen boardPen = new Pen(Brushes.Black, 5);
        private Brush FirstPlayer = Brushes.Red,
            SecondPlayer = Brushes.Blue;
        private Color background = Color.DarkGoldenrod;

        #endregion

        private PictureBox box;
        private Graphics g;
        private Game game;
        private int pieceSize;
        private refreshMethod rm;

        private delegate void refreshMethod();

        public Drawer(Game game, PictureBox b)
        {
            this.box = b;
            this.game = game;

            this.box.Image = new Bitmap(box.Width, box.Height);
            g = Graphics.FromImage(box.Image);
            this.pieceSize = Math.Min(box.Width / game.SizeX, box.Height / game.SizeY);
            rm += box.Refresh;
        }

        private void drawGrid()
        {
            g.DrawLine(boardPen, game.SizeX * pieceSize, 0, game.SizeX * pieceSize, game.SizeY * pieceSize);
            g.DrawLine(boardPen, 0, game.SizeY * pieceSize, game.SizeX * pieceSize, game.SizeY * pieceSize);

            for (int i = 0; i < game.SizeX; i++)
            {
                g.DrawLine(Pens.Black, i * pieceSize, 0, i * pieceSize, game.SizeY * pieceSize);
            }

            for (int i = 0; i < game.SizeY; i++)
            {
                g.DrawLine(Pens.Black, 0, i * pieceSize, game.SizeX * pieceSize, i * pieceSize);
            }
        }

        public void DrawGameState()
        {
            this.pieceSize = Math.Min(box.Width / game.SizeX, box.Height / game.SizeY);
            g.Clear(background);
            int[] triplets = AlphaBetaEngine.GetTripletsCount(game);
            int[] bonuses = AlphaBetaEngine.GetPositionBonuses(game);
			int score = triplets[0] * 100 + bonuses[0] - triplets[1] * 100 - bonuses[1];
            drawGrid();
            g.DrawString((triplets[0] * 100 + bonuses[0]) + "\t" + (triplets[1] * 100 + bonuses[1]) + "\t" + (score > 0 ? "+" : "") + score, new Font("Arial", 12), Brushes.Black, 0, 0);
            for (int i = 0; i < game.SizeX; i++)
            {
                for (int j = 0; j < game.SizeY; j++)
                {
                    if (game.Board[i, j] != null)
                    {
                        Brush b = game.Board[i, j] == true ? FirstPlayer : SecondPlayer;
                        g.FillEllipse(b, i * pieceSize, (game.SizeY - j - 1) * pieceSize, pieceSize, pieceSize);
                    }
                }
            }
            for (int i = game.SizeY - 1; i >= 0; i--)
            {
                if (game.Board[game.LastMove, i] != null)
                {
                    g.FillEllipse(Brushes.Black, (game.LastMove + 0.25f) * pieceSize, (game.SizeY - i - 0.75f) * pieceSize, pieceSize/2, pieceSize/2);
                    break;
                }
            }
            box.Invoke(rm);
        }

        public int HandleClick(int coordX, int coordY)
        {
            return coordX / pieceSize;
        }
    }
}
