﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace FourInARow
{
    class Game
    {
        private Random r;

        public Dictionary<bool, Dictionary<int, int>> PiecesPositions;

        public int LastMove = -1;

        public int SizeX
        {
            get;
            private set;
        }
        public int SizeY
        {
            get;
            private set;
        }

        public bool?[,] Board
        {
            get;
            private set;
        }

        public bool HasFinished
        {
            get;
            private set;
        }
        public bool? Winner = null;

        public bool PlayerToMove = true;

        private int[] freeTiles;
        public List<int> PossibleMoves
        {
            get;
            private set;
        }

        public void PlayMove(bool isFirstPlayer, int column)
        {
            if(column < 0 || column >= SizeX)
                return;
            PiecesPositions[isFirstPlayer][column]++;
            for (int i = 0; i < SizeY; i++)
            {
                if (Board[column, i] == null)
                {
                    Board[column, i] = isFirstPlayer;
                    freeTiles[column]--;
                    if (freeTiles[column] == 0)
                        PossibleMoves.Remove(column);
                    checkFinished(column, i);
                    PlayerToMove = !isFirstPlayer;
                    LastMove = column;
                    return;
                }
            }
        }

        private void checkFinished(int column, int row)
        {
            if (PossibleMoves.Count == 0)
            {
                HasFinished = true;
                return;
            }

            bool? side = Board[column, row];
            for (int i = 0; i < SizeX; i++)
            {
                for (int j = 0; j < SizeY; j++)
                {
                    if (Board[i, j] != side)
                        continue;
                    for (int k = 0; k < 4; k++)
                    {
                        if (j + k >= SizeY || Board[i, j + k] != side)
                            break;
                        if (k == 3)
                        {
                            HasFinished = true;
                            Winner = side;
                            return;
                        }
                    }

                    for (int k = 0; k < 4; k++)
                    {
                        if (j + k >= SizeY || i + k >= SizeX || Board[i + k, j + k] != side)
                            break;
                        if (k == 3)
                        {
                            HasFinished = true;
                            Winner = side;
                            return;
                        }
                    }

                    for (int k = 0; k < 4; k++)
                    {
                        if (i + k >= SizeX || Board[i + k, j] != side)
                            break;
                        if (k == 3)
                        {
                            HasFinished = true;
                            Winner = side;
                            return;
                        }
                    }

                    for (int k = 0; k < 4; k++)
                    {
                        if (j - k < 0 || i + k >= SizeX || Board[i + k, j - k] != side)
                            break;
                        if (k == 3)
                        {
                            HasFinished = true;
                            Winner = side;
                            return;
                        }
                    } 
                }
            }
        }

        public List<Action> GetPossibleActions()
        {
            List<Action> result = new List<Action>();
            foreach (var item in PossibleMoves)
            {
                result.Add(new Action(item, PlayerToMove));
            }
            return result;
        }

        public void Reset()
        {
            this.Board = new bool?[SizeX, SizeY];
            this.HasFinished = false;
            freeTiles = new int[SizeX];
            PossibleMoves = new List<int>();
            this.PiecesPositions = new Dictionary<bool, Dictionary<int, int>>();
            this.PiecesPositions[true] = new Dictionary<int, int>();
            this.PiecesPositions[false] = new Dictionary<int, int>();
            for (int i = 0; i < SizeX; i++)
            {
                freeTiles[i] = SizeY;
                PossibleMoves.Add(i);
                this.PiecesPositions[true].Add(i, 0);
                this.PiecesPositions[false].Add(i, 0);
            }
            PlayerToMove = true;
            Winner = null;
        }

        public Game(int sizeX, int sizeY)
        {
            this.r = Program.r;
            this.SizeX = sizeX;
            this.SizeY = sizeY;
            Reset();
        }

        private bool isWinningMove(int column)
        {
            if (column < 0 || column >= SizeX)
                return false;
            for (int i = 0; i < SizeY; i++)
            {
                if (Board[column, i] == null)
                {
                    Board[column, i] = PlayerToMove;
                    checkFinished(column, i);
                    bool copy = HasFinished;
                    HasFinished = false;
                    Board[column, i] = null;
                    return HasFinished;
                }
            }
            return false;
        }

        public void PlayRandomMove()
        {
            /*
            foreach (var item in possibleMoves)
            {
                if (isWinningMove(item))
                {
                    playMove(playerToMove, item);
                    return;
                }
            }
             */
            PlayMove(PlayerToMove, PossibleMoves[r.Next(PossibleMoves.Count)]);
        }

        public Game Clone()
        {
            Game g = new Game(this.SizeX, SizeY);
            g.Winner = this.Winner;
            g.PossibleMoves = new List<int>(this.PossibleMoves);
            g.HasFinished = this.HasFinished;

            this.freeTiles.CopyTo(g.freeTiles, 0);

            for (int i = 0; i < SizeX; i++)
            {
                for (int j = 0; j < SizeY; j++)
                {
                    g.Board[i, j] = this.Board[i, j];
                }   
            }
            return g;
        }

        public void UndoMove(int move)
        {
            int i = SizeY - 1; ;
            while (Board[move, i] == null)
                i--;
            PiecesPositions[(bool)Board[move, i]][move]--;
            Board[move, i] = null;
            if (freeTiles[move] == 0)
                this.PossibleMoves.Insert(0, move);
            freeTiles[move]++;
            this.HasFinished = false;
            this.LastMove = -1;
            this.PlayerToMove = !PlayerToMove;
            
        }
    
    }
}























