﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.Windows.Forms;
using System.Threading;

namespace FourInARow
{
    class GameHandler
    {
        private Player first, second;
        private Game g;
        private Drawer d;
        Task<int> playFirst, playSecond;
        PictureBox pb;

        public void Restart()
        {
            g.Reset();
            first.Reset();
            second.Reset();
        }

        public GameHandler(Player first, Player second, int sizeX, int sizeY, PictureBox box)
        {
            this.first = first;
            this.second = second;
            g = new Game(sizeX, sizeY);
            this.d = new Drawer(g, box);
            playFirst = new Task<int>(first.SelectMove);
            playSecond = new Task<int>(second.SelectMove);
            this.pb = box;
        }

        public void HandleResize()
        {
            d = new Drawer(g, pb);
            d.DrawGameState();
        }

        public void Start(System.Action finishedCallback)
        {
            while (!g.HasFinished)
            {
                playFirst = new Task<int>(first.SelectMove);
                playSecond = new Task<int>(second.SelectMove);

                playFirst.Start();
                    
                int move = playFirst.Result;

                if (!g.PossibleMoves.Contains(move))
                    throw new Exception();

                first.PlayMove(true, move);
                second.PlayMove(false, move);
                g.PlayMove(true, move);
                d.DrawGameState();

                if (g.HasFinished)
                {
                    showResults();
                    break;
                }

                //-----
                playSecond.Start();
                move = playSecond.Result;

                if (!g.PossibleMoves.Contains(move))
                    throw new Exception();

                first.PlayMove(false, move);
                second.PlayMove(true, move);
                g.PlayMove(false, move);
                d.DrawGameState();

                if (g.HasFinished)
                {
                    showResults();
                    break;
                }
            }
            finishedCallback();
        }

        private void showResults()
        {
            string message = "Red: " + first.GetName() + "\nBlue: " + second.GetName() + "\n";
                
            if (g.Winner == null)
                message += "\nDraw!";
            else
                message += (g.Winner == true ? "\nRed" : "\nBlue") + " wins!";

            MessageBox.Show(message, "Game over");
        }

        public void HandleClick(int coordX, int coordY)
        {
            int column = d.HandleClick(coordX, coordY);
            if (g.PlayerToMove)
                first.HandleClick(column);
            else second.HandleClick(column);
        }
    
    }
}
