﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace FourInARow
{
    class MCEngine : Player
    {
        private int simulationsCount;
        private Game g;
        private Dictionary<int, double> scores;
        private Dictionary<int, int> visitedCounts;
        private Random r;

        public override int SelectMove()
        {
            scores.Clear();
            visitedCounts.Clear();
            foreach (var item in g.PossibleMoves)
            {
                scores.Add(item, 0);
                visitedCounts.Add(item, 0);
            }

            for (int i = 0; i < simulationsCount; i++)
            {
                Game simulGame = g.Clone();
                int move = simulGame.PossibleMoves[r.Next(simulGame.PossibleMoves.Count)];
                simulGame.PlayMove(simulGame.PlayerToMove, move);

                while (!simulGame.HasFinished)
                {
                    simulGame.PlayRandomMove();
                }
                switch(simulGame.Winner)
                {
                    case null:
                        scores[move] += 0.5;
                        break;
                    case true:
                        scores[move] += 1;
                        break;
                    case false: 
                        //nothing here
                        break;
                }
                visitedCounts[move]++;
            }
            double bestScore = -1;
            int bestMove = -1;
            string s = "";
            foreach (var item in scores.Keys)
            {
                double score = scores[item] / visitedCounts[item];
                s += " cislo: " + item + " score: " + score;
                if (score > bestScore)
                {
                    bestScore = score;
                    bestMove = item;
                }
            }
            //MessageBox.Show(s);
            return bestMove;
        }

        public override void PlayMove(bool isThisPlayer, int column)
        {
            g.PlayMove(isThisPlayer, column);
        }

        public override string GetName()
        {
            return "MC engine. " + simulationsCount + " simulations.";
        }

        public MCEngine(int simulationsCount, int sizeX, int sizeY, Random r)
        {
            this.simulationsCount = simulationsCount;
            this.visitedCounts = new Dictionary<int, int>();
            this.scores = new Dictionary<int, double>();
            this.g = new Game(sizeX, sizeY);
            this.r = r;
        }

        public override void Reset()
        {
            g.Reset();
            this.visitedCounts = new Dictionary<int, int>();
            this.scores = new Dictionary<int, double>();
            base.Reset();
        }

        public override void HandleClick(int column)
        {
            //nothing here
        }
    }
}
