﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace FourInARow
{
    class Action
    {
        public int RowNumber;
        public bool IsFirstPlayer;

        public Action(int number, bool player)
        {
            this.RowNumber = number;
            this.IsFirstPlayer = player;
        }
    }

    class MCTreeNode
    {
        public static readonly double desiredBias = 0.9;
        private static readonly int minSuccesorPruningLimit = 4, minVisitedPruningLimit = 50;
        private static double realBias = 0.2;
        public static int visitedBeforeExpansion = 10;
        private static double minScore = 1, maxScore = 0;
        private MCTreeNode parrent;
        public List<MCTreeNode> succesors;
        private bool isLeaf, isFinished = false;
        public Action action; //action that leads from position represented by parrent node to this position
        public Game position;
        public double scoreSum, nVisited;
        public int subtreeSize = 0, subtreeDepth = 0;

        public void UpdateSubtreeDepth(int newDepth)
        {
            if (newDepth > subtreeDepth)
            {
                //desiredBias *= 1.001;
                //desiredBias *= 0.9999;
                subtreeDepth = newDepth;
                if (parrent != null)
                    parrent.UpdateSubtreeDepth(subtreeDepth + 1);
            }
            //else
                //desiredBias *= 1.001;
        }

        public int RecalculateSubtreeSize()
        {
            if (isLeaf)
                return 1;
            int size = 0;
            foreach (var item in succesors)
            {
                size += item.RecalculateSubtreeSize();
            }
            size = size + 1;
            this.subtreeSize = size;
            return size;
        }

        public void UpdateSubtreeSize(int additionalNodesCount)
        {
            this.subtreeSize += additionalNodesCount;
            if (this.parrent != null)
                parrent.UpdateSubtreeSize(additionalNodesCount);
        }

        private MCTreeNode(MCTreeNode parrent)
        {
            this.parrent = parrent;
            this.scoreSum = 0;
            this.nVisited = 1;
            succesors = new List<MCTreeNode>();
        }

        private MCTreeNode()
        {
        }

        public void Expand()
        {
            if (isFinished)
                return;
            if (position.HasFinished)
            {
                isFinished = true;
                return;
            }
            this.isLeaf = false;
            List<Action> possibleActions = position.GetPossibleActions();
            if (possibleActions.Count == 0)
            {
                throw new Exception();
            }
            foreach (var item in possibleActions)
            {
                MCTreeNode succesor = CreateLeaf(this, item);
                succesor.position = this.position.Clone();
                succesor.position.PlayMove(item.IsFirstPlayer, item.RowNumber);
                succesors.Add(succesor);
                if (succesor.position == null)
                {
                    throw new Exception();
                }
            }
            this.position = null;
            this.UpdateSubtreeSize(possibleActions.Count);
            this.UpdateSubtreeDepth(1);
        }

        private void updateBias(double score)
        {
            //desiredBias *= 0.9999;
            if (score < minScore)
                minScore = score;
            if (score > maxScore)
                maxScore = score;
            realBias = desiredBias * (maxScore - minScore) + minScore;
        }

        public void Update(double score)
        {
            this.nVisited++;
            this.scoreSum += score;
            if (this.parrent != null)
                parrent.Update(score);
            else
                updateBias(score);
        }

        public double Eval()
        {
            double expectation = scoreSum / nVisited;
            double urgency = realBias * Math.Sqrt(Math.Log(parrent.nVisited) / nVisited);
            return expectation + urgency;
        }

        private double getExpectation()
        {
            return scoreSum / nVisited;
        }

        public void Prune()
        {
            if (nVisited < minVisitedPruningLimit)
                return;
            int worst = 0, best = 0;
            for (int i = 0; i < succesors.Count; i++)
            {
                if (succesors[i].getExpectation() < succesors[worst].getExpectation())
                    worst = i;
                if (succesors[i].getExpectation() > succesors[best].getExpectation())
                    best = i;
                succesors[i].Prune();
            }
            //succesors[best].prune();
            if (succesors.Count > minSuccesorPruningLimit)
                succesors.RemoveAt(worst);
        }

        public MCTreeNode SelectSuccesor()
        {
            if (this.isLeaf)
                return this;
            double max = -10;
            MCTreeNode bestSuccesor = null;
            foreach (var item in succesors)
            {
                if (item.isFinished)
                {
                    continue;
                }
                double val = item.Eval();
                if (val > max)
                {
                    max = val;
                    bestSuccesor = item;
                }
            }
            if (bestSuccesor == null)
            {
                this.isFinished = true;
                bestSuccesor = succesors[0];
            }
            return bestSuccesor.SelectSuccesor();
        }

        #region static methods
        
        public static MCTreeNode CreateRoot(Game position)
        {
            MCTreeNode node = new MCTreeNode();
            node.succesors = new List<MCTreeNode>();
            node.parrent = null;
            node.scoreSum = 0;
            node.nVisited = 1;
            node.isLeaf = true;
            node.position = position;
            return node;
        }

        public static MCTreeNode CreateLeaf(MCTreeNode parrent, Action move)
        {
            MCTreeNode node = new MCTreeNode(parrent);
            node.isLeaf = true;
            node.action = move;
            return node;
        }

        #endregion


    }
}
